# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .gui import *
from .page import GuiPage
from arjuna.interact.gui.auto.finder.wmd import GuiWidgetMetaData
from arjuna.tpi.tracker import track
from arjuna.tpi.log import *

@track("debug")
class _App(Gui, metaclass=abc.ABCMeta):

    def __init__(self, *, config=None, ext_config=None, label=None, gns_dir=None, gns_file_name=None):
        gns_dir = gns_dir is not None and gns_dir or ""
        super().__init__(gns_dir=gns_dir, config=config, ext_config=ext_config, label=label)
        self.__ui = None
        self.__automator = None
        self.__gns_file_name = gns_file_name is not None and gns_file_name or "{}.yaml".format(self.label)
        self.__common_gui_def = None

    @property
    def _automator(self):
        return self.__automator

    def _launchautomator(self):
        # Default Gui automation engine is Selenium
        from arjuna.interact.gui.auto.automator import GuiAutomator
        self.__automator = GuiAutomator(self, self.config, self.ext_config)

    @property
    def ui(self):
        return self.__ui

    @ui.setter
    def ui(self, page):
        self.__ui = page

    @property
    def _common_guidef(self):
        return self.__common_gui_def

    def _create_default_ui(self):
        self.__ui = GuiPage(source_gui=self, label=self.label, gns_dir=self.gns_dir, gns_file_name=self.__gns_file_name)
        self.__common_gui_def = self.__ui._gui_def

    @abc.abstractmethod
    def launch(self):
        pass

    def prepare_widget(self, widget_object):
        return widget_object

@track("debug")
class GuiApp(_App):
    '''
        Represents a GUI App.

        It supports all methods of **GuiAppContent** by delegating to a default **GuiPage** which it encapsulates.

        Args:
            *args: Any number of positional argumnts. These are passed to the **prepare()** method if defined in inherited class.

        Keyword Arguments:
            url: Base URL for Web application. If not provided, value of **ArjunaOption.APP_URL** is taken from the associated **Configuration** object.
            config: Configuration object.
            ext_config: (Not Supported Yet) AutomatorExtendedConfig object for underlying GUI automator.
            label: Label for the GuiApp. If not provided, the class name is used as the label.
            gns_dir: Relative Root Directory for GNS file(s) associated with the GuiApp. Default is **<ProjectRootDirectory>/guiauto/namespace**. If provided, it is considered relative to the namespace directory.
            gns_file_name: Name of GNS file associated with GuiApp. If not provided, default is **<label>.yaml**.
            kwargs: Arbitrary keyword arugments. These are passed to the **prepare()** method if defined in inherited class.

        Note:
            GuiApp can be used as singular represenation of complete page.

            Through its advanced Python implementation, **GuiApp** supports all methods that are supported by a **GuiPage** by delegating the calls to an internal page object.

            So, although not directly visible in the API docs for GuiApp, you can call all **GuiPage** methods on **GuiApp** as well and they will be executed on the current page in the browser.
    '''

    def __init__(self, *args, url: str=None, config: 'Configuration'=None, ext_config: 'AutomatorExtendedConfig'=None, label: str=None, gns_dir: str=None, gns_file_name: str=None, **kwargs):
        super().__init__(gns_dir=gns_dir, gns_file_name=gns_file_name, config=config, ext_config=ext_config, label=label is None and self.__class__.__name__ or label)
        from arjuna.tpi.constant import ArjunaOption
        self.__url = url is not None and url or self.config.value(ArjunaOption.APP_URL)
        # self._load(*args, **kwargs)
        self.__args = args
        self.__kwargs = kwargs

    @property
    def url(self):
        '''
            Base URL of this GuiApp
        '''
        return self.__url

    def launch(self, *, blank_slate=False):
        '''
            Launch GuiApp.

            Keyword Arguments:
                blank_state: If True, App's base URL is not opened. Default is False.
        '''
        self._launchautomator()
        if self.config.value(ArjunaOption.BROWSER_NETWORK_RECORDER_AUTOMATIC):
            self._automator.network_recorder.record(title=self.__class__.__name__ + " - Home")
        if not blank_slate:
            self._automator.browser.go_to_url(self.url)
        self._create_default_ui()
        self._load(*self.__args, **self.__kwargs)

    def quit(self):
        '''
            Close browser and do cleanup of GuiApp.
        '''
        try:
            if self.config.value(ArjunaOption.BROWSER_NETWORK_RECORDER_AUTOMATIC):
                self._automator.network_recorder.register()
        except:
            pass
        self._automator.quit()

    def __getattr__(self, name):
        return getattr(self.ui, name)

    @property
    def main_window(self):
        '''
           Main window of this **GuiApp**
        '''
        return self._automator.main_window

    def child_window(self, *str_or_with_locators):
        '''
            (Not Supported Yet) Child Window.
        '''
        return self._automator.child_window(self.convert_to_with_wmd(*str_or_with_locators))

    @property
    def latest_child_window(self):
        '''
            (Not Supported Yet) Latest Child Window.
        '''
        return self._automator.latest_child_window

    def close_all_child_windows(self):
        '''
            (Not Supported Yet) Close All Child Windows
        '''
        self._automator.close_all_child_windows()


import copy
import re
from arjuna.tpi.helper.arjtype import CIStringDict
from arjuna.configure.validator import Validator

def _format(target, repl_dict):

    def get_global_value(in_str):
        from arjuna import C, L, R
        gtype, query = in_str.split(".", 1)
        gtype = gtype.upper()
        return locals()[gtype](query)

    pattern = r"\$(\s*[\w\.]*?\s*)\$"
    fmt_target = target.replace("\{", "__LB__").replace("}", "__RB__")

    # Find params
    matches = re.findall(pattern, fmt_target)
    
    for match in matches:
        names_set = None
        target = "${}$".format(match)
        processed_name = match.lower().strip()
        repl_value = None
        if processed_name.startswith("c.") or processed_name.startswith("l.") or processed_name.startswith("r."):
            repl_value = get_global_value(processed_name)
        else:
            if processed_name not in repl_dict:
                continue
            repl_value = repl_dict[processed_name]

        fmt_target = fmt_target.replace(target, str(repl_value))
        

    fmt_target = fmt_target.replace("__LB__", "{").replace("__RB__", "}")
    return fmt_target

class WithX:

    def __init__(self, xdict={}):
        self.__xdict = CIStringDict()
        for k,v in xdict.items():
            try:
                self.__xdict[Validator.name(k)] = {"wtype" : xdict[k]["wtype"].strip().upper(), "wvalue" : xdict[k]["wvalue"]}
            except Exception as e:
                raise Exception(f"Invalid WithX entry for name >>{k}<<.")

    def has_locator(self, name):
        return name in self.__xdict

    def format_args(self, name, vargs, kwargs):
        if not self.has_locator(name):
            raise Exception("No WithX locator with name {} found.".format(name))

        # Critical to create a copy
        fmt = copy.deepcopy(self.__xdict[name])
        repl_dict = {k.lower():v for k,v in kwargs.items()}
        try:
            if fmt["wtype"] in {'ATTR', 'FATTR', 'BATTR', 'EATTR'}: #:, 'NODE', 'BNODE', 'FNODE'}:
                if len(fmt["wvalue"]) > 1:
                    raise Exception("attr/fattr/battr/eattr specification should contain only a single key value pair for attribute name and value")
                name = list(fmt["wvalue"].keys())[0]
                value = list(fmt["wvalue"].values())[0]
                fmt["wvalue"] = {'name' : name, 'value' : value}

            if fmt["wtype"] in {'ATTR', 'FATTR', 'BATTR', 'EATTR', 'NODE', 'BNODE', 'FNODE'}:
                out = dict()
                for k,v in fmt["wvalue"].items():
                    if type(v) in {list, tuple}:
                        out[_format(k, repl_dict)] = [_format(v_entry, repl_dict) for v_entry in v]
                    else:
                        out[_format(k, repl_dict)] = _format(v, repl_dict)
                    #out[k.format(**kwargs)] = v.format(**kwargs)
                return fmt["wtype"], out
            else:
                return fmt["wtype"], _format(["wvalue"], repl_dict) #fmt["wvalue"].format(*vargs, **kwargs)
        except Exception as e:
            from arjuna import log_error
            log_error(f"Error in processing withx {name} : {fmt} for vargs {vargs} and kwargs {kwargs}")
            raise

    def format(self, name, loc_obj):
        vargs = None
        kwargs = None
        if type(loc_obj) is str:
            vargs = [loc_obj]
            kwargs = dict()
        elif type(loc_obj) in {list, tuple}:
            vargs = loc_obj
            kwargs = dict()
        elif type(loc_obj) is dict:
            vargs = []
            kwargs = loc_obj
        return self.format_args(name, vargs, kwargs)
        

    


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class ROProxy:
    def __init__(self, wrapped):
        vars(self)['__wrapped'] = wrapped

    def __getattr__(self, item):
        try:
            return getattr(vars(self)['__wrapped'], item)
        except AttributeError as e:
            raise Exception("Map does not contain attribute: >>{}<<".format(item))

    def __getitem__(self, item):
        try:
            return vars(self)['__wrapped'][item]
        except AttributeError as e:
            raise Exception("Map does not contain key: >>{}<<".format(item))

    def __setattr__(self, key, value):
        raise Exception("Read-Only Proxy does not support item assignment.")

    def __str__(self):
        return str(vars(self)['__wrapped'])




# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from arjuna.tpi.helper.arjtype import CIStringDict

class PipeLine:

    def __init__(self):
        self.__queue = list()

    def pop(self):
        try:
            return self.__queue.pop(0)
        except IndexError as e:
            raise Exception("Pipe line is empty.")

    def push(self, *values):
        self.__queue.extend(values)

    @property
    def length(self):
        return len(self.__queue)

class PipeLines(CIStringDict):

    def __init__(self):
        super().__init__()

    def create_pipeline(self, name):
        self[name] = PipeLine()
        return self[name]

class SharedObjects(CIStringDict):

    def __init__(self):
        super().__init__()

    def create_object(self, name, obj):
        self[name] = obj

class DataStore:

    def __init__(self):
        self.__store = {
            "pipelines" : PipeLines(),
            "shared_objects" : SharedObjects()
        }

    @property
    def pipelines(self):
        return self.__store["pipelines"]

    @property
    def shared_objects(self):
        return self.__store["shared_objects"]

    def create_pipeline(self, name):
        return self.pipelines.create_pipeline(name)

    def get_pipeline(self, name):
        return self.pipelines[name]

    def store_shared_object(self, name, obj):
        self.shared_objects.create_object(name, obj)

    def get_shared_object(self, name):
        return self.shared_objects[name]


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import sys
import types
import inspect
import functools

from arjuna.core.utils.obj_utils import get_class_for_method

class _InvokerFilter(logging.Filter):

    def filter(self, record):
        if not hasattr(record, "invoker"):
            record.invoker = '<no_trace>'
        from arjuna import Arjuna
        if hasattr(record, "contexts"):
            if not record.contexts.intersection(Arjuna._get_allowed_log_contexts()):
                return False

        return True

class Logger:

    def __init__(self, ref_config):
        self.__ref_config = ref_config
        self.__add_trace_level()
        self.__logger = None
        self.__load()

    def __load(self):
        from arjuna.tpi.constant import ArjunaOption
        dl = logging.getLevelName(self.__ref_config.value(ArjunaOption.LOG_CONSOLE_LEVEL).name) # logging.getLevelName(
        log_dir = self.__ref_config.value(ArjunaOption.LOG_DIR)
        if not os.path.isdir(log_dir):
            os.makedirs(log_dir)
        fl = logging.getLevelName(self.__ref_config.value(ArjunaOption.LOG_FILE_LEVEL).name)
        fname = "arjuna.log"
        lpath = os.path.join(log_dir, fname)

        logger = logging.getLogger("arjuna")
        logger.addFilter(_InvokerFilter())
        logger.setLevel(logging.TRACE)
        ch = logging.StreamHandler(sys.stdout)
        ch.flush = sys.stdout.flush
        ch.setLevel(dl)
        fh = logging.FileHandler(lpath, "w", 'utf-8')
        fh.setLevel(fl)
        #f_fmt = logging.Formatter(u'[%(levelname)5s]\t%(asctime)s\t%(pathname)s::%(module)s.%(funcName)s:%(lineno)d\t%(message)s')
        f_fmt = logging.Formatter(u'[%(levelname)7s]\t%(asctime)s\t%(invoker)s\t%(message)s')
        c_fmt = logging.Formatter(u'[LOG] %(message)s')
        ch.setFormatter(c_fmt)
        fh.setFormatter(f_fmt)
        logger.addHandler(ch)
        logger.addHandler(fh)

        self.__logger = logger

    def __add_trace_level(self):
        logging.TRACE = logging.DEBUG - 5
        logging.addLevelName(logging.TRACE, "TRACE")
        def trace(self, message, *args, **kws):
            if self.isEnabledFor(logging.TRACE):
                # Yes, logger takes its '*args' as 'args'.
                self._log(logging.TRACE, message, args, **kws)
        logging.Logger.trace = trace

    @property
    def arjuna_logger(self):
        return self.__logger

'''
    def __load_console(self, dl, logger):
        class __console:
            lock = threading.RLock()

            def __init__(self):
                # self.lock = threading.RLock()
                self.separator = os.linesep
                self.log_display_level = dl
                self.logger = logger

            @sync_method('lock')
            def __log(self, message, err=False):
                if err:
                    mparts = message.replace(u"\r\n", u"--|--").replace(u"\n", u"--|--").split(u'--|--')
                    for mpart in mparts:
                        self.logger.error(mpart)
                else:
                    mparts = message.replace(u"\r\n", u"--|--").replace(u"\n", u"--|--").split(u'--|--')
                    for mpart in mparts:
                        self.logger.info(mpart)

            @sync_method('lock')
            def __print(self, message):
                print(message, end='', file=sys.stdout, flush=True)

            @sync_method('lock')
            def __eprint(self, message):
                print(message, end='', file=sys.stderr, flush=True)

            @sync_method('lock')
            def __println(self, message):
                print(message, file=sys.stdout, flush=True)

            @sync_method('lock')
            def __eprintln(self, message):
                print(message, file=sys.stderr, flush=True)

            @sync_method('lock')
            def __msg(self, *messages):
                return " ".join([str(m) for m in messages])

            @sync_method('lock')
            def display(self, *messages):
                message = self.__msg(*messages)
                should_print = self.__log(message)
                if should_print:
                    self.__println(message)

            @sync_method('lock')
            def display_error(self, *messages):
                message = self.__msg(*messages)
                should_print = self.__log(message, err=True)
                if should_print:
                    self.__eprintln(message)

            @sync_method('lock')
            def error_for_console(self, *messages):
                message = self.__msg(*messages)
                self.__eprintln(message)

            @sync_method('lock')
            def display_on_same_line(self, *messages):
                message = self.__msg(*messages)
                should_print = self.__log(message)
                if should_print:
                    self.__print(message)

            @sync_method('lock')
            def marker(self, length, symbol='-'):
                self.display(symbol * length)

            @sync_method('lock')
            def marker_error(self, length, symbol='-'):
                self.display_error(symbol * length)

            @sync_method('lock')
            def marker_on_same_line(self, length=40):
                self.display_on_same_line("-" * length)

            @sync_method('lock')
            def display_key_value(self, key, value):
                message = "%s %s".format(key, value)
                self.display(message)

            @sync_method('lock')
            def __get_formatted_key_value(self, key, value, left_padding):
                if not left_padding:
                    message = "| {:20s}| {}".format(key, value)
                else:
                    message = "| {}| {}".format(key.ljust(left_padding), value)
                return message

            @sync_method('lock')
            def __display_paddedKV(self, key, value, print_func, left_padding):
                print_func(self.__get_formatted_key_value(key, value, left_padding))

            @sync_method('lock')
            def display_padded_key_value(self, key, value, left_padding=None):
                self.__display_paddedKV(key, value, self.display, left_padding)

            @sync_method('lock')
            def display_padded_key_value_error(self, key, value, left_padding=None):
                self.__display_paddedKV(key, value, self.display_error, left_padding)

            @sync_method('lock')
            def display_exception_block(self, e, strace):
                self.marker_error(80)
                self.display_padded_key_value_error("Exception Type", e.__class__.__name__, 30)
                self.display_padded_key_value_error("Exception Message", str(e), 30)
                self.display_padded_key_value_exception_trace("Exception Trace", strace, 30)
                self.marker_error(80)

            def set_central_log_level(self, level):
                self.central_log_level = level

            @sync_method('lock')
            def display_multiline_key_value(self, key, value, left_padding=30):
                value = str(value)
                ctrace_parts = value.replace("\t", " ").replace("\r\n\r\n","\r\n").replace("\n\n","\n").split(sys_utils.get_line_separator())

                header = self.__get_formatted_key_value(key, ctrace_parts[0], left_padding)
                self.__log(header)
                # if should_print:
                #     self.error_for_console(header)

                for s in ctrace_parts[1:]:
                    message = self.__get_formatted_key_value("", s, left_padding)
                    self.__log(message)
                    # if should_print:
                    #     self.(message)

            display_padded_key_value_exception_trace = display_multiline_key_value
        self.__console = __console()
'''


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import os
from enum import Enum
import functools

from arjuna.interact.gui.auto.finder.wmd import GuiWidgetMetaData
from arjuna.tpi.helper.arjtype import Dictable

from arjuna.interact.gui.gom.guidef import *
from arjuna.tpi.guiauto.meta.formatter import GuiWidgetLocatorFormatter
from arjuna.tpi.engine.asserter import AsserterMixIn

from arjuna.core.poller.conditions import *
from arjuna.core.poller.caller import *
from arjuna.tpi.error import *
from arjuna.core.error import *
from arjuna.tpi.guiauto.model.gns import GNS
from arjuna.interact.gui.auto.finder import GuiFinder, GuiEmdFinder

from .gui import Gui
from arjuna.tpi.helper.image import Image
from arjuna.tpi.protocol.screen_shooter import ScreenShooter
from arjuna.tpi.tracker import track

@track("debug")
class GuiAppContent(Gui, ScreenShooter):
    '''
        Represents content of any type in a **GuiApp**.

        This is the base class for **GuiPage**, **GuiSection** and **GuiDialog** and wraps the underlying **GuiAutomator** object.

        Args:
            *args: Any number of positional argumnts. These are passed to the **prepare()** method if defined in inherited class.

        Keyword Arguments:
            automator: (Mandatory) GuiAutomator object as created by the **GuiApp**.
            label: Label for the this **GuiAppContent**. If not provided, the class name is used as the label.
            gns_dir: Relative Root Directory for GNS file(s) associated with this **GuiAppContent**. Default is the GNS directory of parent **GuiApp**.
            gns_file_name: Name of GNS file associated with this **GuiAppContent**. If not provided, default is **<label>.yaml**.
            **kwargs: Arbitrary keyword arugments. These are passed to the **prepare()** method if defined in inherited class.
    '''

    def __init__(self, *args, automator: 'GuiAutomator', label: str=None, gns_dir: str=None, gns_file_name: str=None, **kwargs):
        self.__app = automator.app
        self.__automator = automator
        gns_dir = gns_dir and gns_dir or self.app.gns_dir
        super().__init__(gns_dir=gns_dir, config=automator.config, ext_config=automator.ext_config, label=label)
        gns_file_name = gns_file_name is not None and gns_file_name or "{}.yaml".format(self.label)
        self.__def_file_path = os.path.join(self.gns_dir, gns_file_name)

        from arjuna import Arjuna
        self.__guimgr = Arjuna.get_gui_mgr()
        self.__guidef = None
        self.__gui_registered = False
        self._externalize()
        self.__gns = None

        self.__finder = GuiFinder(self)
        self.__wmd_finder = GuiEmdFinder(self)
        
    def _load_gns(self):
        self.__gns = GNS(self, self._gui_def)

    @property
    def network_recorder(self):
        return self.__automator.network_recorder

    def send_keys(self, key_chord):
        from arjuna.tpi.guiauto.meta.locator import GuiWidgetLocator
        locator = GuiWidgetLocator(tag="body")
        body = self.locate(locator)
        body.send_keys(key_chord)

    @property
    def gns(self):
        '''
            GNS object associated with this **GuiAppContent** object.

            Encapsulates the externlized namespace and all GuiWidget meta data inside it.
        '''
        return self.__gns

    @property
    def _finder(self):
        return self.__finder

    @property
    def _wmd_finder(self):
        return self.__wmd_finder

    @property
    def app(self):
        '''
            **GuiApp** associated with this **GuiAppContent**.
        '''
        return self.__app

    @property
    def _automator(self):
        return self.__automator

    @property
    def _gui_def(self):
        return self.__guidef

    def _externalize(self):
        try:
            self.__guidef = GuiDef(self.__guimgr.name_store, self._automator, self.label, self.def_file_path)
        except Exception as e:
            import traceback
            raise GuiNamespaceLoadingError(self, str(e) + traceback.format_exc())

        from arjuna import log_debug
        log_debug("Gui Namespace loading completed for {}.".format(self.label))

    @property
    def def_file_path(self):
        '''
            Absolute GNS File path associated with this **GuiAppContent**.
        '''
        return self.__def_file_path

    def _transit(self, page):
        pass

    @property
    def _root_element(self):
        return None

    @property
    def browser(self):
        '''
            **Browser** object in which this **GuiAppContent** exists.
        '''
        return self._automator.browser

    def formatter(self, **fargs) -> GuiWidgetLocatorFormatter:
        '''
            Create a :class:`~arjuna.tpi.guiauto.meta.formatter.GuiWidgetLocatorFormatter` object.

            Keyword Arguments:
                **fargs: Arbitrary key-value pairs to be used for formatting identifiers in `GuiWidgetLocator`.
        '''
        return GuiWidgetLocatorFormatter(self, **fargs)

    def _wait_until_absent(self, wmd):
        try:
            self._automator._wait_until_element_absent(wmd)
        except ArjunaTimeoutError:
            raise GuiWidgetPresentError(self, wmd)         

    def wait_until_absent(self, *, fargs=None, **kwargs):
        '''
            Wait until a **GuiWidget** is absent.

            Keyword Arguments:
                fargs: A dictionary of key-value pairs for formatting the `GuiWidgetLocator`. Use **.format(**kwargs).wait_until_absent** for more Pythonic code when formatting.
                **kwargs: Arbitrary key-value pairs used to construct a `GuiWidgetLocator`

            Note:
                By default Wait is done until **ArjunaOption.GUIAUTO_MAX_WAIT** in the **Configuration** object associated with this **GuiAppContent**.

                You can pass **max_wait** argument to change this. Value is considered in seconds.
        '''
        from arjuna.tpi.guiauto.meta.locator import GuiWidgetLocator
        wmd = GuiWidgetLocator(fmt_args=fargs, **kwargs)._as_wmd()
        self._wait_until_absent(wmd)

    def contains(self, *, fargs=None, **kwargs):
        '''
            Check whether this **GuiAppContent** object contains a **GuiWidget**. Includes dynamic waiting.

            Keyword Arguments:
                fargs: A dictionary of key-value pairs for formatting the `GuiWidgetLocator`. Use **.format(**kwargs).wait_until_absent** for more Pythonic code when formatting.
                **kwargs: Arbitrary key-value pairs used to construct a `GuiWidgetLocator`

            Note:
                By default Wait is done until **ArjunaOption.GUIAUTO_MAX_WAIT** in the **Configuration** object associated with this **GuiAppContent**.

                You can pass **max_wait** argument to change this. Value is considered in seconds.
        '''
        try:
            self.element(fargs=fargs, **kwargs)
        except GuiWidgetNotPresentError:
            return False
        else:
            return True

    @property
    def _dom_root(self):
        '''
            **GuiApp** associated with this **GuiAppContent**.
        '''
        return self._automator.dom_root(self)

    def _frame(self, *str_or_with_locators):
        return self._automator.frame(self, self.convert_to_with_wmd(*str_or_with_locators))

    @property
    def _alert(self):
        return self._automator.alert

    @property
    def title(self):
        '''
            (Not Supported Yet) Title of the window containing this **GuiAppContent**.
        '''
        return self.app.main_window.title

    def set_slomo(self, *, on, interval=None):
        '''
            (Not Supported Yet) Set Slow Motion mode.

            Keyword Arguments:
                on: If True the mode is switched on.
                interval: Number of seconds between successive actions.
        '''
        self._automator.set_slomo(on, interval)

    def execute_javascript(self, js, *args):
        '''
            Inject and Execute JavaScript.

            Args:
                js: Arbitrary JavaScript
                *args: Any number of positional arguments used for formatting of JavaScript by underlying automation engine.
        '''
        return self._automator.execute_javascript(js, *args)

    def _take_screenshot(self, prefix: str=None) -> Image:
        return self._automator.take_screenshot(prefix=prefix)

    def go_to_url(self, url: str=None):
        '''
            Go to a URL. 

            This action takes place in the current browser window/tab.
        '''
        self.browser.go_to_url(url)

    def _load_anchor_element(self):
        label = self._gui_def.anchor_element_name

        from arjuna import log_debug
        log_debug("Loading Anchor Element for {} Gui. anchor label in GNS: {}.".format(
            self.label,
            self._gui_def.anchor_element_name,
        ))

        if label is not None:
            getattr(self.gns, label)

    def locate(self, locator) -> 'GuiWidget':
        '''
           Locate a `GuiWidget`.

           Arguments:
            locator: `GuiWidgetLocator` object.

            Returns:
                An object of type `GuiWidget`. Exact object type depends on the value of **type** attribute in `GuiWidgetLocator`. 
        '''
        return self._finder.locate(locator)

    def element(self, *, fargs=None, **kwargs):
        '''
            Locate a `GuiElement`.

            Keyword Arguments:
                fargs: A dictionary of key-value pairs for formatting the `GuiWidgetLocator`. Use **.format(kwargs).wait_until_absent** for more Pythonic code when formatting.
                **kwargs: Arbitrary key-value pairs used to construct a `GuiWidgetLocator`

            Returns:
                `GuiElement` object.
        '''
        return self._finder.element(fargs=fargs, **kwargs)

    def multi_element(self, fargs=None, **kwargs):
        '''
            Locate a `GuiMultiElement`.

            Keyword Arguments:
                fargs: A dictionary of key-value pairs for formatting the `GuiWidgetLocator`. Use **.format(kwargs).wait_until_absent** for more Pythonic code when formatting.
                **kwargs: Arbitrary key-value pairs used to construct a `GuiWidgetLocator`

            Returns:
                `GuiMultiElement` object.
        '''
        return self._finder.multi_element(fargs=fargs, **kwargs)

    def dropdown(self, fargs=None, **kwargs):
        '''
            Locate a `GuiDropDown`.

            Keyword Arguments:
                fargs: A dictionary of key-value pairs for formatting the `GuiWidgetLocator`. Use **.format(kwargs).wait_until_absent** for more Pythonic code when formatting.
                **kwargs: Arbitrary key-value pairs used to construct a `GuiWidgetLocator`

            Returns:
                `GuiDropDown` object.
        '''
        return self._finder.dropdown(fargs=fargs, **kwargs)

    def radio_group(self, fargs=None, **kwargs):
        '''
            Locate a `GuiRadioGroup`

            Keyword Arguments:
                fargs: A dictionary of key-value pairs for formatting the `GuiWidgetLocator`. Use **.format(kwargs).wait_until_absent** for more Pythonic code when formatting.
                **kwargs: Arbitrary key-value pairs used to construct a `GuiWidgetLocator`

            Returns:
                `GuiRadioGroup` object
        '''
        return self._finder.radio_group(fargs=fargs, **kwargs)





# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest

class Asserter:
    '''
        Arjuna's asserter class.

        It can be used directly. It is already included as an attribute for request fixture in test functions and all Guis and Gui Templates.
    '''

    def __init__(self):
        self.__asserter = unittest.TestCase('__init__')

    @classmethod
    def _format_msg(cls, msg):
        return msg and " {}.".format(msg) or ""

    def assert_equal(self, obj1, obj2, msg):
        '''
            Assert obj1 == obj2

            Wrapper on unittest's assertEqual

            Args:
                obj1: Object of any type which supports == operator for obj2 type
                obj2: Object meeting the above constraint
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertEqual(obj1, obj2, msg)

    def assert_lesser(self, obj1, obj2, msg):
        '''
            Assert obj1 < obj2

            Wrapper on unittest's assertLess

            Args:
                obj1: Object of any type which supports < operator for obj2 type
                obj2: Object meeting the above constraint
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertLess(obj1, obj2, msg)

    def assert_greater(self, obj1, obj2, msg):
        '''
            Assert obj1 > obj2

            Wrapper on unittest's assertLess

            Args:
                obj1: Object of any type which supports > operator for obj2 type
                obj2: Object meeting the above constraint
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertGreater(obj1, obj2, msg)

    def assert_min(self, obj, min_value, msg):
        '''
            Asserts a minimum value for an object i.e. obj >= min_value

            Wrapper on unittest's assertGreaterEqual

            Args:
                obj: Object of any type which supports >= operator for min_value
                min_value: Object meeting the above constraint
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertGreaterEqual(obj, min_value, msg)

    def assert_max(self, obj, max_value, msg):
        '''
            Asserts a maximum value for an object i.e. obj <= max_value

            Wrapper on unittest's assertLessEqual

            Args:
                obj: Object of any type which supports <= operator for min_value
                max_value: Object meeting the above constraint
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertLessEqual(obj, max_value, msg)

    def assert_not_equal(self, obj1, obj2, msg):
        '''
            Assert obj1 != obj2

            Wrapper on unittest's assertNotEqual

            Args:
                obj1: Object of any type which supports != operator for obj2 type
                obj2: Object meeting the above constraint
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertNotEqual(obj1, obj2, msg)

    def assert_true(self, obj, msg):
        '''
            Assert obj is True.

            Wrapper on unittest's assertTrue

            Args:
                obj: Object of any type
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertTrue(obj, msg)

    def assert_false(self, obj, msg):
        '''
            Assert obj is False.

            Wrapper on unittest's assertFalse

            Args:
                obj: Object of any type
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertFalse(obj, msg)

    def fail(self, msg):
        '''
            Raises AssertionError with the provided message.

            Args:
                msg: A context string explaining the failure.
        '''
        self.__asserter.fail(msg)


class AsserterMixIn:
    '''
        Base class to add **asserter** property to any class which inherits from it.
    '''

    def __init__(self):
        # Trick to use assertions outside of a unittest test
        self.__asserter = Asserter()

    @property
    def asserter(self) -> Asserter:
        '''
            Arjuna's Asserter object for executing assertions.
        '''
        return self.__asserter

class IterableAsserterMixin:

    '''
        Base class for adding assertions to iterables.

        Note:
            Use it in combination with AsserterMixIn.
            The child class should have `_container` as its attribute or property.
    '''

    def __init__(self):
        self.__klass_name = self.__class__.__name__

    def __len__(self):
        return len(self._container)

    def is_empty(self):
        '''
            Returns True if this iterable has no objects.
        '''
        return len(self) == 0

    def assert_empty(self, *, msg):
        '''
            Assert that there should not be any object in this iterable.

            Keyword Arguments:
                msg: Purpose of this assertion.
        '''
        if not self.is_empty():
            raise AssertionError("{} is not empty. Length: {}. {}".format(self.__klass_name, len(self), msg))

    def assert_not_empty(self, *, msg):
        '''
            Assert that there should be atleast one object in this iterable.

            Keyword Arguments:
                msg: Purpose of this assertion.
        '''
        if self.is_empty():
            raise AssertionError(f"{self.__klass_name} is empty. {msg}")

    def assert_size(self, size, *, msg):
        '''
            Assert that the number of objects in this iterable should match provided size.

            Arguments:
                size: Expected number of objects.
            
            Keyword Arguments:
                msg: Purpose of this assertion.
        '''
        length = len(self)
        if length != size:
            raise AssertionError(f"{self.__klass_name} is not of expected size. Expected: {size}. Actual: {length}. {msg}")

    def assert_min_size(self, size, *, msg):
        '''
            Assert that the number of objects in this iterable >= provided size.

            Arguments:
                size: Expected minimum number of objects.
            
            Keyword Arguments:
                msg: Purpose of this assertion.
        '''
        length = len(self)
        if length < size:
            raise AssertionError(f"{self.__klass_name} is not of minimum expected size. Expected minimum size is {size}. Actual: {length}. {msg}")

    def assert_max_size(self, size, *, msg):
        '''
            Assert that the number of objects in this iterable <= provided size.

            Arguments:
                size: Expected maximum number of objects.
            
            Keyword Arguments:
                msg: Purpose of this assertion.
        '''
        if size <0 :
            raise TypeError("Negative number aren't allowed")
        length = len(self)
        if length > size:
            raise AssertionError(f"{self.__klass_name} is not of maximum expected size. Expected maximum size is {size}. Actual: {length}. {msg}")

    def assert_size_range(self, min_size, max_size, *, msg):
        '''
            Assert that the number of objects in this iterable should be in the provided size range.

            Arguments:
                min_size: Expected minimum number of objects.
                max_size: Expected maximum number of objects.
            
            Keyword Arguments:
                msg: Purpose of this assertion.
        '''
        length = len(self)
        if length < min_size or length > max_size:
            raise AssertionError(f"{self.__klass_name} is not in expected size range. Expected maximum size is {size}. Actual: {length}. {msg}")







# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
# import multiprocessing

from arjuna.core.error import TestGroupsFinished
from signal import signal, SIGINT

class TestGroupRunner(threading.Thread):

    def __init__(self, nprefix, wnum, commands):
        super().__init__(name="{}t{}".format(nprefix, wnum))
        self.__commands =  commands
        self.daemon = True # The process is killed when Arjuna exits.

    @property
    def commands(self):
        return self.__commands

    # def handle_interrupt(self, signum, frame):
    #     self.terminate()

    def run(self):
        # signal(SIGINT, self.handle_interrupt)
        from arjuna import log_info
        print("Group runner started")
        while True:
            try:
                child = self.__commands.next()
            except TestGroupsFinished as e:
                print("Groups finished")
                return
            except Exception as e:
                print("An exception occured in thread pooling. Would continue executing.")
                print(e)
                import traceback
                traceback.print_exc()
                return
            
            try:
                print("Running Group Thread")
                child.thread_name = self.name
                child.run()
            except Exception as e:
                import traceback
                print(str(e) + traceback.format_exc())
                continue
        print("Group runner started")



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import abc

from arjuna.core.reader.excel import *
from arjuna.tpi.data.record import *
from arjuna.tpi.parser.yaml import Yaml

from arjuna.tpi.data.reference import IndexedDataReference, ContextualDataReference

class ExcelIndexedDataReference(IndexedDataReference):

    def __init__(self, path):
        self.path = path
        self.__name = get_file_name(path)
        if (path.lower().endswith("xls")):
            self.reader = ExcelRow2MapReader(path)
        else:
            raise Exception("Unsupported file format for Excel reading.")

        map = dict()
        for index, record in enumerate(self.reader):
            map[index] = DataRecord(context="Ref-{}[{}]".format(self.__name, index), **record)
        self.reader.close()
        super().__init__(map)

class YamlIndexedDataReference(IndexedDataReference):

    def __init__(self, path):
        self.path = path
        self.__name = get_file_name(path)
        if (path.lower().endswith("yaml")):
            self.__node = Yaml.from_file(path)
        else:
            raise Exception("Unsupported file format for Excel reading.")

        map = dict()
        for index, record in enumerate(self.__node):
            map[index] = DataRecord(context="Ref-{}[{}]".format(self.__name, index), **dict(record.items()))
        super().__init__(map)

def get_file_name(path):
    name = os.path.basename(path)
    return os.path.splitext(name)[0]

class YamlContextualDataReference(ContextualDataReference):

    def __init__(self, path):
        if (path.lower().endswith("yaml")):
            self.__node = Yaml.from_file(path)
        else:
            raise Exception("Unsupported file format for Excel reading.")
        super().__init__(path)

    def _populate(self):
        for context in self.__node:
            self._add_record(context, dict(self.__node.get_section(context).items()))


class __ExcelDataReference(ContextualDataReference):
    def __init__(self, path):
        if (path.lower().endswith("xls")):
            self.reader = ExcelRow2ArrayReader(path)
        else:
            raise Exception("Unsupported file format for Excel reading.")
        super().__init__(path)

    @abc.abstractmethod
    def _populate(self):
        pass

class ExcelContextualDataReference(__ExcelDataReference):

    def __init__(self, path):
        super().__init__(path)

    def _populate(self):
        contexts = self.reader.headers[1:]
        cmap = {i: {} for i in contexts}
        while True:
            try:
                data_record = self.reader.next()
            except StopIteration:
                break
            else:
                name = data_record[0]
                for index, context in enumerate(contexts):
                    cmap[context][name] = data_record[index+1]
        self.reader.close()
        for context, value in cmap.items():
            self._add_record(context, value)

def R(query="", *, bucket=None, context=None, index=None):
    if context is not None and index is not None:
        raise Exception("You can either specify context (for contextual references) or index (for indexed references).")
    from arjuna import Arjuna, ArjunaOption
    bucket = bucket
    context = context
    index = index
    query = query

    final_query = ""

    bucket = bucket is not None and bucket.lower() + "." or "."
    context_or_index = None
    if context is not None:
        context_or_index = context.lower() + "."
    elif index is not None:
        context_or_index = str(index) + "."
    else:
        context_or_index = "."

    final_query = bucket + context_or_index + query.lower().strip()
    final_query = final_query.replace("..",".")
    if final_query.startswith('.'):
        final_query = final_query[1:]

    if final_query.endswith('.'):
        final_query = final_query[:-1]

    try:
        bucket, context_or_index, query = final_query.split('.', 2)
    except ValueError:
        query = ""
        try:
            bucket, context_or_index = final_query.split('.', 1)
        except:
            context_or_index = ""
            try:
                bucket = final_query
            except:
                raise Exception("Not able to form a valid reference query with provided data. Invalid query: {}".format(final_query))

    try:
        context_or_index = int(context_or_index)
    except:
        if context_or_index == "":
            context_or_index = None

    try:
        query = int(query)
    except:
        if query == "":
            query = None

    try:
        if context_or_index is not None:
            if query is None:
                return Arjuna.get_data_ref(bucket)._record_for(context_or_index)
            else:
                return Arjuna.get_data_ref(bucket)._record_for(context_or_index)[query]
        else:
            return Arjuna.get_data_ref(bucket)
    except Exception as e:
        import traceback
        raise Exception("Error in retrieving reference value for: bucket: >>{}<<, context: >>{}<< and query >>{}<< in data reference. {}. {}".format(bucket, context, query, str(e), traceback.format_exc()))


    # if bucket is not None:
    #     bucket = bucket.lower()
    #     if context is not None:
    #         context = context.lower()
    #         query = query.lower()
    #     else:
    #         if query.find('.') != -1:
    #             context, query = query.split('.', 1)
    #             context = context.lower()
    #             query = query.lower()
    #         else:
    #             raise Exception("DataRefError: The query must specify context using dot notation when not passed as argument.")
    # else:
    #     if context is not None:
    #         raise Exception("bucket must be provided if context arg is passed.")
    #     else:
    #         if query.find('.') != -1:
    #             bucket, context, query = query.split('.', 2)
    #             bucket = bucket.lower()
    #             context = context.lower()
    #             query = query.lower()
    #         else:
    #             raise Exception("DataRefError: The query must specify bucket and context using dot notation when these are not passed as arguments.")           






# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Classes to assist in JSON Parsing. 
'''

import json
import copy
import abc
from arjuna.tpi.helper.arjtype import _ArDict
from jsonpath_rw import jsonpath
from jsonpath_rw_ext import parse
from json.decoder import JSONDecodeError
from genson import SchemaBuilder
from typing import Any

from arjuna.tpi.tracker import track
from arjuna.tpi.engine.asserter import AsserterMixIn, IterableAsserterMixin

@track("trace")
class JsonElement(AsserterMixIn, IterableAsserterMixin, metaclass=abc.ABCMeta):
    '''
        Abstract Json Element. Base class for JsonList and JsonDict.
    '''

    def __init__(self, container):
        AsserterMixIn.__init__(self)
        IterableAsserterMixin.__init__(self)
        self._container = container

    def findall(self, query) -> list:
        '''
            Find all elements with JsonPath query.

            Arguments:
                query: JsonPath Query string

            Returns:
                python list of found Json elements.
        '''

        jsonpath_expr = parse(query)
        return [Json.from_object(match.value, allow_any=True) for match in jsonpath_expr.find(self._container)]


    def find(self, query) -> Any:
        '''
            Find first element with JsonPath.

            Arguments:
                query: JsonPath Query string

            Returns:
                A Json element.
        '''
        results = self.findall(query)
        if not results:
            raise Exception(f"A Json element could not be found with {query}")
        return results[0]


@track("trace")
class JsonList(JsonElement):
    '''
        Encapsulates a list object in Json.

        Arguments:
            pylist: Python list

        Note:
            Supports indexing just like a Python list.

            Also supports == operator. Right operand can be a Python list or a `JsonList` object.
    '''

    def __init__(self, pylist: list):
        self.__list = pylist
        if pylist is None:
            self.__list = list()
        super().__init__(self.__list)

    def __getitem__(self, index):
        return Json.from_object(self.__list[index], allow_any=True)

    def __len__(self):
        return len(self.__list)

    @property
    def size(self):
        '''
            Number of objects in this list.
        '''
        return len(self)

    @property
    def raw_object(self):
        '''
            A copy of the Python list that this object wraps.
        '''
        return copy.deepcopy(self.__list)

    def __str__(self):
        return str(self.__list)

    def __len__(self):
        return len(self.__list)

    def __eq__(self, other):
        if type(other) is list:
            pass
        elif type(other) is JsonList:
            other = other.raw_object
        elif other is None:
            return False
        else:
            raise Exception("JsonList == operator expects list/JsonList as right operand.")

        if len(self) != len(other):
            return False
        else:
            return self.raw_object == other


@track("trace")
class JsonDict(_ArDict, JsonElement):
    '''
        Encapsulates Dictionary object in Json.

        Arguments:
            pydict: Python dict.

        Note:
            Supports dictionary methods as well as **.** access for key.

            Also supports == operator. Right operand can be a Python dict or a `JsonDict` object.

    '''

    def __init__(self, pydict: dict=None):
        _ArDict.__init__(self, pydict)
        JsonElement.__init__(self, self.store)

    @property
    def size(self):
        '''
            Number of key-value pairs in this object.
        '''
        return len(self)

    @property
    def raw_object(self):
        '''
            A copy of the Python dict that this object wraps.
        '''
        return dict(super().items())

    def _process_key(self, key: str):
        return key

    def __getitem__(self, query):
        return self.find(query)

    def assert_contents(self, *, msg, **key_values):
        '''
            Assert that the values of provided keys match in the root of this dict.

            Arguments:
                msg: Purpose of this assertion.
                **key_values: Arbitrary key-value pairs to match in the root of this dict.

        '''
        for k,v in key_values.items():
            self.asserter.assert_equal(Json.from_object(self[k], allow_any=True), v, msg=msg)

    def assert_keys_present(self, *keys, msg):
        '''
            Assert that the provided keys are present in the root of this dict.

            Arguments:
                msg: Purpose of this assertion.
                **keys: Key names to assert
        '''
        for key in keys:
            self.asserter.assert_true(key in self, "Expected key absent. " + msg)

    def matches_schema(self, schema):
        '''
            Check if this JsonDict matches the provided schema

            Arguments:
                schema: Python dict schema or `JsonSchema` object.

            Returns:
                A 2-element tuple (True/False, List of error strings in matching)
        '''
        if isinstance(schema, JsonSchema):
            schema = schema.as_dict()
        from jsonschema import Draft7Validator
        validator = Draft7Validator(schema)
        errors = []
        for error in sorted(validator.iter_errors(self.store), key=str):
            errors.append("Key >>{}<<: {}".format(".".join(error.path), error.message))
        if errors:
            return False, errors
        else:
            return True, None

    def assert_schema(self, schema, *, msg):
        '''
            Assert that this JsonDict matches the provided schema.

            Arguments:
                schema: Python dict schema or `JsonSchema` object.

            Keyword Args:
                msg: Purpose of this assertion.
        '''
        matched, errors = self.matches_schema(schema)
        if not matched:
            raise AssertionError("JsonDict does not match schema. Found the following errors. {}".format(". ".join(errors)))

    def assert_match_schema(self, jobj, *, msg):
        '''
            Assert that this JsonDict matches the schema of provided object.

            Arguments:
                jobj: Json string or Python dict/JsonDict

            Keyword Args:
                msg: Purpose of this assertion.
        '''
        matched, errors = self.matches_schema(Json.extract_schema(jobj))
        if not matched:
            raise AssertionError("JsonDict does not match schema. Found the following errors. {}".format(". ".join(errors)))

    def assert_match(self, jobj, *, msg, ignore_keys=None):
        '''
            Assert that this JsonDict matches the provided object.

            Arguments:
                jobj: Python dict/JsonDict

            Keyword Args:
                msg: Purpose of this assertion.
                ignore_keys: Keys to be ignored while matching
        '''
        if type(jobj) is dict:
            jobj = JsonDict(jobj)
        elif type(jobj) is JsonDict:
            pass
        else:
            raise Exception("JsonDict.assert_match expected_dict argument should be a dictionary or a JsonDict object.")
        if ignore_keys is None:
            ignore_keys = set()
        elif type(ignore_keys) in {set, dict, list, tuple}:
            ignore_keys = set([i.lower() for i in ignore_keys])
        else:
            ignore_keys = {str(ignore_keys).lower()}
        for k,v in self.items():
            if k.lower() not in ignore_keys:
                self.asserter.assert_equal(Json.from_object(v, allow_any=True), jobj[k], msg=msg)

    @property
    def schema(self) -> 'JsonSchema':
        '''
            Schema of this object.
        '''
        return Json.extract_schema(self)

    def __eq__(self, other):
        if type(other) is dict:
            pass
        elif type(other) is JsonDict:
            other = other.raw_object
        elif other is None:
            return False
        else:
            raise Exception("JsonDict == operator expects dict/JsonDict as right operand. Provided {} of type {}".format(other, type(other)))

        if len(self) != len(other):
            return False
        else:
            return self.raw_object == other

@track("trace")
class JsonSchemaBuilder:
    '''
        JsonSchema builder class.

        Keyword Arguments:
            schema_uri: URI to be set for '$schema' key.
    '''

    def __init__(self, *, schema_uri=None):
        if schema_uri:
            self.__builder = SchemaBuilder(schema_uri=schema_uri)
        else:
            self.__builder = SchemaBuilder()

    def schema(self, schema):
        '''
            Add schema to the overall schema.

            Call multiple times to add more schema.

            Arguments:
                schema: Json schema dict
        '''
        self.__builder.add_schema(schema)
        return self

    def string(self, jstr):
        '''
            Add schema based on an object string to the overall schema.

            Call multiple times to add more objects.

            Arguments:
                jstr: A string representing a Python object.
        '''
        jobj = Json.from_str(jstr.strip(), allow_any=True)
        self.object(jobj)
        return self

    def object(self, jobj):
        '''
            Add schema based on an object to the overall schema.

            Call multiple times to add more objects.

            Arguments:
                jobj_or_str: JsonDict/JsonList or a Python str/list/dict.
        '''
        if type(jobj) in {JsonDict, JsonList}:
            self.__builder.add_object(jobj.raw_object)
        else:
            self.__builder.add_object(jobj)
        return self

    def build(self):
        '''
            Create `JsonSchema` based on the building constructs provided till this point.
        '''
        return JsonSchema(self.__builder.to_schema())

@track("trace")
class JsonSchema:
    '''
        Encapsulates Json Schema which can be used to validate Json objects.
    '''

    def __init__(self, schema_dict):
        self.__dict = schema_dict

    def set_maximum(self, number):
        '''
            Set maximum occurances.

            Arguments:
                number: Upper limit for occurances.
        '''
        self.__dict['maximum'] = number

    def set_allowed_values(self, *values):
        '''
            Set allowed values.

            Arguments:
                *values: Arbitrary allowed values.
        '''
        self.__dict['enum'] = values

    def mark_optional(self, *keys):
        '''
            Mark the provided keys as optional in the root of Json Schema.

            Arguments:
                *keys: Arbitrary key names
        '''

        for key in keys:
            if 'required' in self.__dict:
                try:
                    self.__dict['required'].remove(key)
                except ValueError:
                    pass

    def allow_null(self, *keys):
        '''
            Allow None as a value type for the provided keys in the root of Json Schema.

            Arguments:
                *keys: Arbitrary key names
        '''

        for key in keys:
            if key in self.__dict['properties']:
                target_dict = self.__dict['properties'][key]
                if 'type' in target_dict:
                    type_value = target_dict['type']
                    if type(type_value) is str:
                        self.__dict['properties'][key]['type'] = [type_value, "null"]
                    else:
                        self.__dict['properties'][key]['type'].append('null')
                elif 'anyOf' in target_dict:
                    if {"type": "null"} not in target_dict['anyOf']:
                        target_dict['anyOf'].append({"type": "null"})                   

    def as_dict(self) -> dict:
        '''
            Get this object as a Python dict.
        '''
        return self.__dict

    def __str__(self):
        return json.dumps(self.__dict, indent=2)

    @classmethod
    def builder(self, *, schema_uri=None):
        '''
            Create a JsonSchemaBuilder.

            Keyword Arguments:
                schema_uri: URI to be set for '$schema' key.
        '''
        return JsonSchemaBuilder(schema_uri=schema_uri)

class Json:
    '''
        Helper class to create Arjuna's Json elements.
    '''

    @classmethod
    def reset_auto_id_key(cls):
        '''
            Set Auto ID Key for `jsonpath` lib to None. Has global impact.
        '''
        jsonpath.auto_id_field = None

    @classmethod
    def set_auto_id_key(cls, key="id"):
        '''
            Set Auto ID Key for `jsonpath` lib to provided key.

            Arguments:
                key: Name of key. Default is **id**
        '''
        jsonpath.auto_id_field = key

    @classmethod
    def from_object(cls, jobj, *, allow_any=False):
        '''
            Convert a Python object to a JsonDict/JsonList object if applicable.

            Any iterable which is not a dict is converted to a JsonList.

            Arguments:
                jobj: dict/list/tuple/set or a compatible Python object.

            Keyword Arguments:
                allow_any: If True, if the object can not be coverted, same object is returned, else an Exception is raised.
        '''
        if type(jobj) is dict:
            return JsonDict(jobj)
        elif type(jobj) in {list, tuple, set}:
            return JsonList(jobj)
        elif type(jobj) is not str:
            try:
                iter(jobj)
                return JsonList(jobj)
            except:
                pass
            if allow_any:
                return jobj
            else:
                raise Exception(f"Not able to convert provided string {jobj} into any appropriate Json element.")
        else:
            # string
            if allow_any:
                return jobj     
            else:
                raise Exception(f"Not able to convert provided string {jobj} into any appropriate Json element.")       

    @classmethod
    def from_str(cls, jstr: str, *, allow_any: bool=False) -> 'JsonDictOrList':
        '''
            Creates a Json object from a JSON string.

            Arguments:
                jstr: A string representing a compatible Python object.

            Keyword Arguments:
                allow_any: If True, if the object can not be coverted to a JsonDict/JsonList, same object is returned, else an Exception is raised.

            Returns:
                Arjuna's `JsonDict` or `JsonList` object or the same object for allow_any = True
        '''
        try:
            jobj = json.loads(jstr)
        except JSONDecodeError:
            if not allow_any:
                raise
            return cls.from_object(jstr, allow_any=allow_any)
        else:
            return cls.from_object(jobj, allow_any=allow_any)


    @classmethod
    def from_file(cls, file_path: str, *, allow_any: bool=False) -> 'JsonDictOrList':
        '''
            Creates a Json object from file.

            Arguments:
                file_path: Absolute path of the json file.

            Keyword Arguments:
                allow_any: If True, if the object can not be coverted to a JsonDict/JsonList, same object is returned, else an Exception is raised.

            Returns:
                Arjuna's `JsonDict` or `JsonList` object or the same object for allow_any = True
        '''

        with open(file_path, 'r') as f:
            return cls.from_str(f.read(), allow_any=allow_any)

    @classmethod
    def from_map(cls, map: dict) -> JsonDict:
        '''
            Creates a JsonDict object from Python dictionary.

            Arguments:
                map: Python dict

            Returns:
                Arjuna's `JsonDict` object
        '''

        return JsonDict(d)

    @classmethod
    def from_iter(cls, iter) -> JsonList:
        '''
            Creates a Json object from Python iterator.

            Arguments:
                iter: Python iterable

            Returns:
                Arjuna's `JsonList` object
        '''

        return JsonList(list(iter))

    @classmethod
    def assert_list_type(cls, obj, *, msg):
        '''
            Assert that the provided object is a Python list or `JsonList` object.

            Keyword Args:
                msg: Purpose of this assertion.
        '''
        if not isinstance(obj, list) and not isinstance(obj, JsonList):
            raise AssertionError("Object {} is not a list. {}".format(str(obj), msg))

    @classmethod
    def assert_dict_type(cls, obj, *, msg):
        '''
            Assert that the provided object is a Python dict or `JsonDict` object.

            Keyword Args:
                msg: Purpose of this assertion.
        '''
        if not isinstance(obj, dict) and not isinstance(obj, JsonDict):
            raise AssertionError("Object {} is not a dict. {}".format(str(obj), msg))

    @classmethod
    def extract_schema(cls, jobj_or_str) -> JsonSchema:
        '''
            Extracts schema from provide Json object or string to create a `JsonSchema` object.

            Arguments:
                jobj_or_str: JsonDict/JsonList or a Python list/dict or a string representing a Python object
        '''
        builder = JsonSchema.builder()
        if type(jobj_or_str) is str:
            builder.string(jobj_or_str)
        else:
            builder.object(jobj_or_str)
        return builder.build()
        

    



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.tpi.constant import ArjunaOption
from arjuna.tpi.parser.yaml import Yaml
from arjuna.core.error import *
from arjuna.tpi.error import *
from .stage import TestStage, MagicTestStage, YamlTestStage, MagicTestStageForGroup
from arjuna.tpi.parser.yaml import YamlList

class BaseTestSession:

    def __init__(self, name, config, dry_run=False):
        self.__name = name
        self.__config = config
        self.__dry_run = dry_run
        self.__stages = []

    def add_stage(self, stage):
        self.__stages.append(stage)

    @property
    def name(self):
        return self.__name

    @property
    def config(self):
        return self.__config

    @property
    def dry_run(self):
        return self.__dry_run
  
    def run(self):
        from arjuna import log_info
        for stage in self.__stages:
            print("Executing stage: {} ...".format(stage.name))
            stage.run()
            print("Finished Executing stage: {} ...".format(stage.name))
            import time
            time.sleep(2)


class MagicTestSession(BaseTestSession):

    def __init__(self, config, *, rules=None, dry_run=False):
        super().__init__("msession", config, dry_run=dry_run)

        stage = MagicTestStage(session=self, rules=rules)
        self.add_stage(stage)

class MagicTestSessionForStage(BaseTestSession):

    def __init__(self, stage_name, config, dry_run=False):
        super().__init__("msession", config, dry_run=dry_run)

        YamlTestSession.load_session_defs()
        self.add_stage(YamlTestStage(name=stage_name, stage_yaml=YamlTestSession.get_stage_yaml(stage_name), session=self))

class MagicTestSessionForGroup(BaseTestSession):

    def __init__(self, group_name, config, dry_run=False):
        super().__init__("msession", config, dry_run=dry_run)

        YamlTestSession.load_session_defs()
        self.add_stage(MagicTestStageForGroup(session=self, group_name=group_name))

class YamlTestSession(BaseTestSession):

    __SESSIONS_YAML = None
    SESSIONS_YAML_FILE = None
    __STAGES_YAML = None
    STAGES_YAML_FILE = None
    __GROUPS_YAML = None
    GROUPS_YAML_FILE = None

    @classmethod
    def __load_sessions_file(cls):
        from arjuna import C
        if cls.__SESSIONS_YAML is None:
            cls.SESSIONS_YAML_FILE = C(ArjunaOption.CONF_SESSIONS_FILE)
            try:
                cls.__SESSIONS_YAML = Yaml.from_file(cls.SESSIONS_YAML_FILE, allow_any=True)
            except FileNotFoundError as e:
                raise TestSessionsFileNotFoundError(file_path=cls.SESSIONS_YAML_FILE)

    @classmethod
    def get_session_yaml(cls, name):
        try:
            return cls.__SESSIONS_YAML.get_section(name)
        except YamlUndefinedSectionError as e:
            raise UndefinedTestSessionError(name=name, file_path=cls.SESSIONS_YAML_FILE)

    @classmethod
    def __load_stages_file(cls):
        from arjuna import C
        if cls.__STAGES_YAML is None:
            cls.STAGES_YAML_FILE = C(ArjunaOption.CONF_STAGES_FILE)
            try:
                cls.__STAGES_YAML = Yaml.from_file(cls.STAGES_YAML_FILE, allow_any=True)
            except FileNotFoundError:
                raise TestStagesFileNotFoundError(file_path=cls.STAGES_YAML_FILE)

    @classmethod
    def get_stage_yaml(cls, name):
        try:
            return cls.__STAGES_YAML.get_section(name)
        except YamlUndefinedSectionError as e:
            raise UndefinedTestStageError(name=name, file_path=cls.STAGES_YAML_FILE)

    @classmethod
    def __load_groups_file(cls):
        from arjuna import C
        if cls.__GROUPS_YAML is None:
            cls.GROUPS_YAML_FILE = C(ArjunaOption.CONF_GROUPS_FILE)
            try:
                cls.__GROUPS_YAML = Yaml.from_file(cls.GROUPS_YAML_FILE, allow_any=True)
            except FileNotFoundError:
                raise TestGroupsFileNotFoundError(file_path=cls.GROUPS_YAML_FILE)

    @classmethod
    def get_group_yaml(cls, name):
        try:
            return cls.__GROUPS_YAML.get_section(name)
        except YamlUndefinedSectionError as e:
            raise UndefinedTestGroupError(name=name, file_path=cls.GROUPS_YAML_FILE)

    @classmethod
    def load_session_defs(cls):
        cls.__load_sessions_file()
        cls.__load_stages_file()
        cls.__load_groups_file()

    def __init__(self, name, ref_config_name, dry_run=False):
        self.__yaml = None
        self.load_session_defs()
        self.__yaml = self.get_session_yaml(name)

        from arjuna import Arjuna
        self.__config = None
        if ref_config_name is None:
            self.__config = Arjuna.get_config()
        else:
            self.__config = Arjuna.get_config(ref_config_name)
        self.__load_metadata()
        super().__init__(name, self.__config, dry_run=dry_run)
        self.__load_stages()

    def __load_metadata(self):
        for section_name in self.__yaml.section_names:
            section_name = section_name.lower()
            if section_name.lower() not in {"groups", "include"}:
                if section_name.lower() == "conf":
                    from arjuna import Arjuna
                    self.__config = Arjuna.get_config(self.__yaml["conf"])

    def __load_stages(self):
        from arjuna import Arjuna

        if "include" not in self.__yaml.section_names:
            raise InvalidTestSessionDefError(
                            session_name=self.name, 
                            sessions_file_path=self.SESSIONS_YAML_FILE, 
                            msg="It must contain 'include' section. Section names found: {}".format(tuple(self.__yaml.section_names))
            )

        for section_name in  self.__yaml.section_names:
            if section_name.lower() == "include":
                stage_names = self.__yaml[section_name]
                if type(stage_names) not in {list, YamlList} or set([type(s) is str for s in stage_names]) != {True}:
                    raise InvalidTestSessionDefError(
                                    session_name=self.name, 
                                    sessions_file_path=self.SESSIONS_YAML_FILE, 
                                    msg="'include' section can only contain a YAML list of stage names. Found:\n{}:\n {}".format(section_name, str(self.__yaml[section_name]))
                    )
                for stage_name in stage_names:
                    self.add_stage(YamlTestStage(name=stage_name, stage_yaml=self.get_stage_yaml(stage_name), session=self))
            # elif section_name.lower() == "stages":
            #     stages_yaml = self.__yaml.get_section(section_name)
            #     stages = stages_yaml.section_names
            #     if not stages:
            #         raise InvalidTestSessionDefError(
            #                         session_name=self.name, 
            #                         sessions_file_path=self.__SESSIONS_YAML.file_path, 
            #                         msg="'stages' section must define atleast one stage."
            #         )
                
            #     for stage in stages:
            #         self.add_stage(YamlTestStage(stage_yaml=stages_yaml.get_section(stage), session=self))



import re
from arjuna.core.constant import GuiElementType

from ._with import GuiGenericLocator
from .enums import GenericLocateWith

class LocatorTranslator:
    '''
        Translates With Constructs to what underlying engines understand.
    '''

    BASIC_LOCATORS = {
        GenericLocateWith.ID,
        GenericLocateWith.NAME,
        GenericLocateWith.XPATH,
        GenericLocateWith.IMAGE,
        GenericLocateWith.INDEX,
        GenericLocateWith.WINDOW_TITLE,
        GenericLocateWith.WINDOW_PTITLE,
        GenericLocateWith.JS,
    }

    NEED_TRANSLATION = {
        GenericLocateWith.LINK : GenericLocateWith.PARTIAL_LINK_TEXT,
        GenericLocateWith.FLINK : GenericLocateWith.LINK_TEXT,
        GenericLocateWith.SELECTOR : GenericLocateWith.CSS_SELECTOR,
    }

    XTYPE_LOCATORS = {
        GuiElementType.TEXTBOX: "//input[@type='text']",
        GuiElementType.PASSWORD: "//input[@type='password']",
        GuiElementType.LINK: "//a",
        GuiElementType.BUTTON: "//input[@type='button']",
        GuiElementType.SUBMIT_BUTTON: "//input[@type='submit']",
        GuiElementType.DROPDOWN: "//select",
        GuiElementType.CHECKBOX: "//input[@type='checkbox']",
        GuiElementType.RADIO: "//input[@type='radio']",
        GuiElementType.IMAGE: "//img",
    }

    XPATH_LOCATORS = {
        GenericLocateWith.TEXT : "//*[contains(text(),'{}')]",
        GenericLocateWith.FTEXT : "//*[text()='{}']",
        GenericLocateWith.BTEXT : "//*[starts-with(text(),'{}')]",
        GenericLocateWith.VALUE : "//*[@value='{}']",
        GenericLocateWith.TITLE : "//*[@title='{}']",
        GenericLocateWith.IMAGE_SRC : "//img[@src='{}']"
    }

    NAMED_ARG_LOCATORS = {
        # GenericLocateWith.ATTR : "//*[contains(@{name},'{value}')]",
        # GenericLocateWith.FATTR : "//*[@{name}='{value}']",
        # GenericLocateWith.BATTR : "//*[starts-with(@{name},'{value}')]",
        # GenericLocateWith.EATTR : "//*[ends-with(@{name},'{value}')]",
        GenericLocateWith.ATTR : "*[{name}*='{value}']",
        GenericLocateWith.FATTR : "*[{name}='{value}']",
        GenericLocateWith.BATTR : "*[{name}^='{value}']",
        GenericLocateWith.EATTR : "*[{name}$='{value}']",
        GenericLocateWith.POINT : "return document.elementFromPoint({x}, {y})",
    }

    TEXT_TRANSLATIONS = {
        'text' : "text()",
        '*text' : "*//text()",
        'star_text' : "*//text()",
        '.text' : ".",
        'dot_text' : ".",
    }

    @classmethod
    def translate(cls, locator):
        from arjuna import log_debug

        def process_selector(cssvalue):
            css_string = None
            if type(cssvalue) is str:
                css_string = "." + cssvalue.replace('.', ' ').strip()
            else:
                if type(cssvalue[0]) is str:
                    # Tuple like ('button button-large',)
                    cssvalue = " ".join(cssvalue)
                    css_string = "." + cssvalue.replace('.', ' ').strip()
                else:
                    # Tuple is (('button', 'button-large'),)
                    css_string = "." + ".".join(cssvalue[0])
            return re.sub(r'\s+', '.', css_string)

        def convert_to_class_list(cssvalue):
            return process_selector(cssvalue)[1:].split(".")

        def process_tags(tagsvalue):
            tag_list = None
            if type(tagsvalue) is str:
                tag_list = tagsvalue.strip().split()
            else:
                tag_list = tagsvalue
            return [t.lower()=='any' and '*' or t for t in tag_list]

        rltype = locator.ltype
        rlvalue = locator.lvalue
        glvalue = None
        try:
            gltype = GenericLocateWith[rltype.upper()]
        except Exception as e:
            raise Exception("Invalid locator across all automators: {}={}. Error: {}".format(rltype, type(rlvalue), str(e)))
        else:
            log_debug("Processing locator: Type: {}; Value: {}".format(str(gltype), rlvalue))
            if gltype == GenericLocateWith.ELEMENT:
                glvalue = rlvalue
            elif gltype in cls.BASIC_LOCATORS:
                glvalue = rlvalue
            elif gltype in cls.NEED_TRANSLATION:
                glvalue = rlvalue
                gltype = cls.NEED_TRANSLATION[gltype]
            elif gltype in cls.XPATH_LOCATORS:
                glvalue = cls.XPATH_LOCATORS[gltype].format(rlvalue)
                gltype = GenericLocateWith.XPATH
            elif gltype == GenericLocateWith.POINT:
                glvalue = cls.NAMED_ARG_LOCATORS[gltype].format(**rlvalue)
            elif gltype in cls.NAMED_ARG_LOCATORS:
                glvalue = cls.NAMED_ARG_LOCATORS[gltype].format(**rlvalue)
                gltype = GenericLocateWith.CSS_SELECTOR
            elif gltype == GenericLocateWith.CLASSES:
                css_string = process_selector(rlvalue)
                # if type(rlvalue) is str:
                #     css_string = "." + rlvalue.replace('.', ' ').strip()
                # else:
                #     if type(rlvalue[0]) is str:
                #         css_string = "." + rlvalue[0].replace('.', ' ').strip()
                #     else:
                #         css_string = "." + ".".join(rlvalue[0])
                glvalue = process_selector(rlvalue)
                gltype = GenericLocateWith.CSS_SELECTOR
            elif gltype == GenericLocateWith.TAGS:
                glvalue = " ".join(process_tags(rlvalue))
                gltype = GenericLocateWith.CSS_SELECTOR                
            elif gltype in {GenericLocateWith.NODE, GenericLocateWith.BNODE, GenericLocateWith.FNODE}:
                lkeys = [k.lower() for k in rlvalue.keys()]
                if 'tag' in lkeys and 'tags' in lkeys:
                    raise Exception("node Locator definition can contain either 'tag' or 'tags' a key.")
                if len(set(lkeys).intersection(set(cls.TEXT_TRANSLATIONS.keys()))) > 1:
                    raise Exception("node Locator definition can contain only one of text/star_text/*text/dot_text/.text keys.")
                # Initial processing
                updated_rlvalue = dict()
                use_xpath = False
                for k,v in rlvalue.items():
                    if k.lower() == 'tag':
                        updated_rlvalue['tags'] = process_tags(v)
                    elif k.lower() == 'tags':
                        updated_rlvalue['tags'] = process_tags(v)
                    elif k.lower() == 'classes':
                        updated_rlvalue['classes'] = convert_to_class_list(v)
                    elif k.lower() in cls.TEXT_TRANSLATIONS:
                        updated_rlvalue[k.lower()] = v
                        use_xpath = True
                    elif k.lower() == "use_xpath":
                        from arjuna.core.types.constants import TRUES, FALSES
                        v = str(v).strip()
                        if v.lower() not in TRUES and v.lower() not in FALSES:
                            raise Exception("Allowed values for use_path are true/false/on/off")
                        if v.lower() in TRUES:
                            use_xpath = True
                    else:
                        updated_rlvalue[k] = v
                rlvalue = updated_rlvalue
                if use_xpath:
                    # Create XPath
                    xblocks = []
                    tags = '*'
                    for k,v in rlvalue.items():
                        if k.lower() == 'tags':
                            tags = "//".join(v)
                            continue

                        if k == "classes":
                            for c in v:
                                xblocks.append(f"contains(@class,'{c}')")
                            continue

                        if k.lower() in cls.TEXT_TRANSLATIONS:
                            k = cls.TEXT_TRANSLATIONS[k.lower()]
                        else:
                            k = f'@{k}'
                        if gltype == GenericLocateWith.NODE:
                            xblocks.append(f"contains({k},'{v}')")
                        elif gltype == GenericLocateWith.BNODE:
                            xblocks.append(f"starts-with({k},'{v}')")
                        elif gltype == GenericLocateWith.FNODE:
                            xblocks.append(f"{k}='{v}'")
                    if xblocks:
                        xblocks_str = "[" + " and ".join(xblocks) + "]"
                    else:
                        xblocks_str = ""
                    glvalue = f"//{tags}{xblocks_str}"
                    gltype = GenericLocateWith.XPATH
                else:
                    cblocks = []
                    tags = '*'
                    classes = ""
                    for k,v in rlvalue.items():
                        if k.lower() == 'tags':
                            tags = " ".join(v)
                            continue

                        if k == "classes":
                            classes = "." + ".".join(v)
                            continue
                        
                        if gltype == GenericLocateWith.NODE:
                            cblocks.append(f"[{k}*='{v}']")
                        elif gltype == GenericLocateWith.BNODE:
                            cblocks.append(f"[{k}^='{v}']")
                        elif gltype == GenericLocateWith.FNODE:
                            cblocks.append(f"[{k}='{v}']")
                    if cblocks:
                        cblocks_str = "".join(cblocks)
                    else:
                        cblocks_str = ""
                    glvalue = f"{tags}{classes}{cblocks_str}"
                    gltype = GenericLocateWith.CSS_SELECTOR
            else:
                raise Exception("Locator not supported yet by Arjuna: " + rltype)
    
        return GuiGenericLocator(gltype, glvalue)


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import inspect
from numbers import Number as PyNumber
from weakref import WeakKeyDictionary
from enum import Enum

from . import constants


def _get_cstr(context=None):
    return context and "Encountered Type error for {}.".format(context) or ""


class TypeValidationError(TypeError):
    pass


class Descriptor:
    pass

class String(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if String.check(input):
            raise TypeValidationError("{} {} [type: {}] is not a string.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        return isinstance(input, str)

class Integer(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if Integer.check(input):
            raise TypeValidationError("{} {} [type: {}] is not a integer.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        return isinstance(input, str)

class Float(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if Float.check(input):
            raise TypeValidationError("{} {} [type: {}] is not a flot.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        return isinstance(input, str)

class Number(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if Number.check(input):
            raise TypeValidationError("{} {} [type: {}] is not a number.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        return isinstance(input, PyNumber)

class Bool(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if Bool.check(input):
            raise TypeValidationError("{} {} [type: {}] is not a bool.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        return isinstance(input, bool)

    @staticmethod
    def force_convert(input):
        if Bool.check(input):
            return input
        else:
            if input in constants.TRUES:
                return True
            elif input in constants.FALSES:
                return False
            else:
                return bool(input)

class Enumeration(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if Enumeration.check(input):
            raise TypeValidationError("{} Provided enum_type: {} [type: {}] is not a class.".format(_get_cstr(context), input, type(input)))
        return input

    @staticmethod
    def check(input):
        if not inspect.isclass(input) or not issubclass(input, Enum):
            return False

class EnumConstant(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if EnumConstant.check(input):
            raise TypeValidationError("{} {} [type: {}] is not a enum.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        return isinstance(input, Enum)

    @staticmethod
    def convert(enum_type, estr, context=None, enum_type_validated=False):
        if not enum_type_validated: Enumeration.validate(enum_type)

        if String.check(estr):
            uestr = estr.upper()
            if uestr in enum_type.__members__:
                return enum_type[uestr]
            else:
                raise TypeError("{} {} is not a valid enum constant for enum {}.".format(_get_cstr(context), estr, enum_type))

class EnumConstantList(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if EnumConstantList.check(input):
            raise TypeValidationError("{} {} [type: {}] is not an enum constant list.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        if not List.check(input): return False
        for i in input:
            if not EnumConstant.check(i):
                return False
        return True

    @staticmethod
    def convert(enum_type, input, context=None):
        Enumeration.validate(enum_type)
        il = List.check(input) and input or [input]
        return [EnumConstant.convert(enum_type, estr, enum_type_validated=True) for estr in il]

class List(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if not List.check(input):
            raise TypeValidationError("{} {} [type: {}] is not a list.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        return isinstance(input, list)

class Tuple(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if not Tuple.check(input):
            raise TypeValidationError("{} {} [type: {}] is not a tuple.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        return isinstance(input, tuple)

class Set(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if not Set.check(input):
            raise TypeValidationError("{} {} [type: {}] is not a set.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        return isinstance(input, set)

class ListTupleOrSet(Descriptor):

    @staticmethod
    def validate(input, context=None):
        if not ListTupleOrSet.check(input):
            raise TypeValidationError("{} {} [type: {}] is not a list, tuple or set.".format(_get_cstr(context)))
        return input

    @staticmethod
    def check(input):
        return List.check(input) or Tuple.check(input) or Set.check(input)

class StringList(Descriptor):
    @staticmethod
    def validate(input, context=None):
        if not isinstance(input, list):
            raise TypeValidationError("{} {} [type: {}] is not a list.".format(_get_cstr(context)))
        else:
            for i in input:
                String.validate(i, "Validating String List. Provided List: {}.".format(str(input)))
        return input

    @staticmethod
    def force_convert(input):
        if input is None:
            return None
        else:
            if String.check(input):
                return [input]
            elif Number.check(input):
                return [str(input)]
            elif ListTupleOrSet.check(input):
                return [str(i) for i in input]
            else:
                raise TypeError("Can not force convert >>{}<< to a string list.".format(str(input)))

class NaturalNumber(Descriptor):
    @staticmethod
    def validate(input, context=None):
        Integer.validate(input, context)
        if input <= 0:
            raise TypeValidationError("{} {} [type: {}] is not a natural number.".format(_get_cstr(context)))
        return input


class RangeBoundInt(Descriptor):
    @staticmethod
    def validate(input, left, right, context=None):
        Integer.validate(input, context)
        if input <= 0:
            raise TypeValidationError("{} {} [type: {}] is not in the range [{}, {}].".format(_get_cstr(context)))
        return input



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from arjuna.tpi.helper.image import Image
from arjuna.tpi.tracker import track


@track("trace")
class NetworkRecorder(metaclass=abc.ABCMeta):

    def __init__(self):
        self.__in_recording_mode = False

    def is_recording(self):
        return self.__in_recording_mode

    @abc.abstractmethod
    def _is_active(self):
        pass

    def record(self, title):
        '''
            Start recording.

            Arguments:
                title: Title representing all captured traffic.
        '''
        if not self._is_active(): return
        if self.is_recording():
            self.register()
        self._record(title=title)
        self.__in_recording_mode = True

    @abc.abstractmethod
    def _record(self, title):
        pass

    def register(self):
        '''
            Register the current recorded traffic with Arjuna.
        '''
        if not self._is_active(): return
        if not self.is_recording(): return
        self._register()
        self.__in_recording_mode = False

    @abc.abstractmethod
    def _register(self):
        pass


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

###############################
# Raised by top layer user API
###############################

'''
Arjuna Exceptions

This module defines Exception classes that represent different types of run-time issues that can happen. The exceptions encapsulate the underlying exception message and stack traces.
'''

import threading

from arjuna.core.error import *
from arjuna.core.utils.obj_utils import get_function_meta_data

class UndefinedConfigError(Exception):
    '''
        Raised when a configurtion is a referenced by its name in code, externalized files or configuration query.
    '''

    def __init__(self, referred_conf_name, defined_conf_names):
        super().__init__("There is no registered configuration for name: {}. Registered: {}".format(referred_conf_name, tuple(defined_conf_names)))

class ConfigCreationError(Exception):
    '''
        Raised when there is an error in the creation of a configuration object.
    '''

    def __init__(self, msg):
        super().__init__(msg)


class GuiWidgetForLabelPresentError(Exception):
    '''
        Raised when a GuiWidget corresponding to a GNS label is unexpectedly found. Involves Dynamic Waiting for absence.
    '''

    def __init__(self, gui, label, message=None):
        message = format_msg(message)
        super().__init__("GuiWidget corresponding to Label {} in Gui {} is present despite waiting.{}".format(label, gui.label, message))


class GuiWidgetPresentError(Exception):
    '''
        Raised when a GuiWidget is unexpectedly found using a GuiWidgetLocator object. Involves Dynamic Waiting for absence.
    '''

    def __init__(self, gui, wmd, message=None):
        message = format_msg(message)
        super().__init__("GuiWidget corresponding to meta data {} in Gui {} is present despite waiting.{}".format(str(wmd), gui.label, message))

class GuiWidgetForLabelNotPresentError(Exception):
    '''
        Raised when a GuiWidget corresponding to a GNS label is NOT found. Involves Dynamic Waiting for presence.
    '''

    def __init__(self, gui, label, message=None):
        message = format_msg(message)
        super().__init__("GuiWidget corresponding to Label {} in Gui {} is NOT present despite waiting.{}".format(label, gui.label, message))


class GuiWidgetNotPresentError(Exception):
    '''
        Raised when a GuiWidget is NOT found using a GuiWidgetLocator object. Involves Dynamic Waiting for presence.
    '''

    def __init__(self, gui, wmd, message=None):
        message = format_msg(message)
        super().__init__("GuiWidget corresponding to meta data {} in Gui {} is NOT present despite waiting.{}".format(str(wmd), gui.label, message))


class GuiNamespaceLoadingError(Exception):
    '''
        Raised when there is a problem with loading a GNS file.
    '''

    def __init__(self, gui, msg):
        message = msg and  " Error message: {}".format(msg) or ""
        super().__init__("Gui namespace was not loaded for >{}<.{}".format(gui.qual_name, message))


class GuiNotLoadedError(WaitableError):
    '''
        Raised when there is a problem at any step of Arjuna's Gui Loading Protocol.
    '''

    def __init__(self, gui, msg):
        message = msg and  " Error message: {}.".format(msg) or ""
        super().__init__("GUI [{}] did not load as expected.{}".format(gui.qual_name, message))


class GuiLabelNotPresentError(Exception):
    '''
        Raised when a non-existing GNS label is referenced.
    '''

    def __init__(self, gns_name, label, context=None, msg=None):
        msg = msg is None and "" or msg
        context_msg = context and  " for context {}".format(context) or ""
        super().__init__("Gui namespace >{}< does not contain element with name: {}{}. {}.".format(gns_name, label, context_msg, msg))

class GuiLabelNotPresentError(Exception):
    '''
        Raised when a non-existing GNS label is referenced.
    '''

    def __init__(self, gns_name, label, context=None, msg=None):
        msg = msg is None and "" or msg
        context_msg = context and  " for context {}".format(context) or ""
        super().__init__(f"No label definition exists for name {label} in {gns_name}, {context} {msg}")


class TestSessionsFileNotFoundError(Exception):
    '''
        Raised when `run-session` CLI sommand is used and sessions.yaml does not exist in `<Project Root Dir>/config` directory.
    '''

    def __init__(self, *, file_path):
        super().__init__(f"Test sessions configuration file does not exist at {file_path}.")


class UndefinedTestSessionError(Exception):
    '''
        Raised when there is no definition for a name used as test session name.
    '''

    def __init__(self, *, name, file_path):
        super().__init__(f"No session definition exists for name >>{name}<< in sessions configuration file at {file_path}.")

class InvalidTestSessionDefError(Exception):
    '''
        Raised when there is an error in the definition of a test session.
    '''

    def __init__(self, *, session_name, sessions_file_path, msg):
        super().__init__(f"Invalid session definition for >>{session_name}<< in sessions file {sessions_file_path}. {msg}")

class TestStagesFileNotFoundError(Exception):
    '''
        Raised when there `run-session/run-stage` CLI sommand is used and stages.yaml is does not exist in `<Project Root Dir>/config` directory.
    '''

    def __init__(self, *, file_path):
        super().__init__(f"Test Stages configuration file does not exist at {file_path}.")

class UndefinedTestStageError(Exception):
    '''
        Raised when there is no definition for a name used as test stage name.
    '''

    def __init__(self, *, name, file_path):
        super().__init__(f"No stage definition exists for name >>{name}<< in stages configuration file at {file_path}")


class InvalidTestStageDefError(Exception):
    '''
        Raised when there is an error in the definition of a test stage.
    '''

    def __init__(self, *, session_name, stage_name, stages_file_path, msg):
        super().__init__(f"Invalid stage definition for >>{stage_name}<< in stages file {stages_file_path}. It is referred in session >>{session_name}<<. {msg}")

class TestGroupsFileNotFoundError(Exception):
    '''
        Raised when there `run-session/run-stage/run-group` CLI sommand is used and `groups.yaml` does not exist at `<Project Root Dir>/config` directory.
    '''

    def __init__(self, *, file_path):
        super().__init__(f"Test groups configuration file does not exist at {file_path}.")

class UndefinedTestGroupError(Exception):
    '''
        Raised when there is no definition for a name used as test group name.
    '''

    def __init__(self, *, name, file_path):
        super().__init__(f"No group definition exists for name >>{name}<< in groups configuration file at {file_path}")

class TestSelectorNotFoundError(Exception):
    '''
        Test Selector object was not found for current thread (and for current test group creation.)

        Every Test Group is run in a thread either sequentially or parallely.

        A Test Selector object is supposed to be created for every test group.

        This exception is raised when Arjuna is not able to find the registered test Selector object for current thread.
    '''

    def __init__(self):
        super().__init__("No test Selector object was found for current test group in thread: {}".format(threading.current_thread().name))


class TestDecoratorError(Exception):
    '''
        Raised when there is an error in @test decorator for a given test function.
    '''

    def __init__(self, func_qual_name, msg):
        super().__init__(f"There is an error in @test decoratior for {func_qual_name}. {msg}")


class HttpUnexpectedStatusCodeError(Exception):
    '''
        Raised when the HTTP status code for a request does not match expected status code(s).
    '''

    def __init__(self, request, response):
        super().__init__(f"Unexpected status code {response.status_code} for {response.url} in {request.method} request.")
        self.__request = request
        self.__response = response

    @property
    def request(self):
        return self.__request

    @property
    def response(self):
        return self.__response

class HttpRequestCreationError(Exception):
    '''
    Raised when there is an error in creating an HTTP request.
    '''

    def __init__(self, msg):
        super().__init__(f"Error in creating HTTP Request. {msg}")

class HttpConnectError(Exception):
    '''
    Raised when there is a connection error.
    '''

    def __init__(self, request, msg):
        super().__init__(f"Connection Error: {request.method} {request.url}. {msg}")
        self.__request = request

    @property
    def request(self):
        return self.__request


class HttpSendError(Exception):
    '''
        Raised when there is an error in sending an HTTP request.
    '''

    def __init__(self, request, response, msg):
        super().__init__(f"Error in HTTP Request: {request.method} {request.url}. {msg}")
        self.__request = request
        self.__response = response

    @property
    def request(self):
        return self.__request

    @property
    def response(self):
        return self.__response

class DisallowedArjunaOptionError(Exception):
    '''
        Raised when there is an attempt to set an ArjunOption value in a configuration stage, where it is not allowed to override the value.
    '''

    def __init__(self, stage, option):
        super().__init__("Option {} is not allowed in a {} configuration.".format(option, stage.name.title()))


class ArjunaOptionValidationError(Exception):

    def __init__(self, option, name, validator):
        super().__init__(
            "Config option value <{}>(type:{}) for <{}> option did not pass the validation check: [{}]".format(
                    option, type(option), name, validator))


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class ClassPassThrough:
    def __init__(self, klass, *cargs, **ckwargs):
        vars(self)['_o'] = klass(*cargs, **ckwargs)

    def __getattr__(self, item):
        if item in ['_o', '__lock__']:
            return vars(self)[item]
        else:
            return self.pass_through_get(item)

    def __setattr__(self, item, value):
        if item in ['_o', '__lock__']:
            vars(self)[item] = value
        else:
            self.pass_through_set(item, value)

    def pass_through_get(self, item):
        return getattr(self._o, item)

    def pass_through_set(self, item, value):
        return setattr(self._o, item, value)

    def __iter__(self):
        return iter(self._o)


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import abc
from urllib.parse import urlparse, urlencode, parse_qs, quote
from requests import Request, Session
from arjuna.tpi.error import HttpUnexpectedStatusCodeError, HttpSendError
from arjuna.tpi.parser.json import Json
from arjuna.tpi.parser.html import Html
from arjuna.tpi.engine.asserter import AsserterMixIn
from requests.exceptions import ConnectionError, TooManyRedirects
import time

class HttpMessage(AsserterMixIn, metaclass=abc.ABCMeta):

    def __init__(self, http_msg):
        super().__init__()
        self.__http_msg = http_msg

    @property
    def headers(self) -> dict:
        ''' 
            HTTP Headers for this message.
        '''
        return self.__http_msg.headers

    def assert_empty_content(self, *, msg):
        '''
            Validates if content is empty.

            Keyword Arguments:
                msg: A context string explaining why this assertion was done.
        '''
        if self.headers['Content-Length'] != '0':
            raise AssertionError("Content is not empty.")

    def assert_non_empty_content(self, *, msg):
        '''
            Validates if content is empty.

            Keyword Arguments:
                msg: A context string explaining why this assertion was done.
        '''
        if self.headers['Content-Length'] == '0':
            raise AssertionError("Content is empty.")

    def assert_header_match(self, header, value=None, *, msg):
        '''
            Validates the presence and optionally the value of an HTTP Header.

            Arguments:
                header: Name of header
                value: Text contained in header. If not provided, then only presence of header is checked.

            Keyword Arguments:
                msg: A context string explaining why this assertion was done.
        '''
        if header not in self.headers:
            raise AssertionError(f">>{header}<< not present in HTTP Request. {msg}")

        if value is not None:
            actual = self.headers[header]
            if actual != value:
                raise AssertionError(f"Value >>header<< is {actual} but was expected to be {value}. {msg}.")

    def assert_header_mismatch(self, header, value=None, *, msg):
        '''
            Validates the absence of header or mismatch of its content with provided value.

            Arguments:
                header: Name of header
                value: Text contained in header. If not provided, then only absence of header is checked.

            Keyword Arguments:
                msg: A context string explaining why this assertion was done.
        '''

        if value is None:
            if header in self.headers:
                raise AssertionError(f">>{header}<< is present in HTTP Request. {msg}")
        else:
            actual = self.headers[header]
            if actual == value:
                raise AssertionError(f"Value >>header<< is {actual}. It was expected to be different. {msg}.")




# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import types
import inspect
import functools

from arjuna.core.utils.obj_utils import get_class_for_method

prop_dict_msg = {
    "fget": ("(Getting Property)","", " Returning: {}"),
    "fset": ("(Setting Property)", ":: Args: {}, Kwargs: {}.", ""),
    "fdel": ("(Deleting Property)","", ""),
}

def trim_arg(arg, max_len=100):
    arg = str(arg)
    return len(arg) > max_len and arg[0:max_len] + "<SNIP>" or arg

def trim_args(args):
    return [trim_arg(arg) for arg in args]

def trim_kwargs(kwargs):
    return {k:trim_arg(v) for k,v in kwargs.items()}

def trim_ret_value(ret):
    return trim_arg(ret, max_len=200)

def func_wrapper(func, level, *vargs, static=False, prop=False, prop_type="fget", **kwargs):
    import arjuna
    from arjuna import log_error
    name = func.__name__
    qualname = func.__qualname__
    level = name.startswith("_") and "trace" or level
    log_call = getattr(arjuna, "log_{}".format(level.strip().lower()))
    if name != qualname and not static:
        pvargs = vargs[1:]
    else:
        pvargs = vargs

    if prop:
        msg_1 = prop_dict_msg[prop_type][0]
        msg_2 = prop_dict_msg[prop_type][1].format(pvargs, kwargs)
        log_call("{} {}{}".format(qualname, msg_1, msg_2))
    else:
        log_call("{}:: Started with args {} and kwargs {}.".format(qualname, trim_args(pvargs), trim_kwargs(kwargs)))
    ret = None
    try:
        ret = func(*vargs, **kwargs)
    except Exception as e:
        import traceback
        log_call("{}:: Exception: {}. Trace: {}".format(qualname, e, traceback.format_exc()))
        raise
    else:
        if prop:
            msg_1 = prop_dict_msg[prop_type][0]
            msg_3 = prop_dict_msg[prop_type][2].format(ret)
            log_call("{}:: Finished.{}".format(qualname, msg_3, msg_3))
        else:
            log_call("{}:: Finished. Returning: {}".format(qualname, trim_ret_value(ret)))
        return ret

def track_func(level="debug", static=False, prop=False, prop_type="fget"):

    def dec(func):
        fname = func.__name__
        if prop is True:
            if not hasattr(func, "_wrapped"):
                func._wrapped = True
            elif func._wrapped:
                return func
        @functools.wraps(func)
        def inner(*vargs, **kwargs):
            return func_wrapper(func, level, *vargs, static=static, prop=prop, prop_type=prop_type, **kwargs)
        return inner

    return dec

def wrap_methods(cls, level): #, *args, **kwargs):
    for attr_name, attr in vars(cls).items():
        if type(attr) is types.FunctionType:
            setattr(cls, attr_name, track_func(level)(attr))
        elif isinstance(attr, classmethod):
            setattr(cls, attr_name, classmethod(track_func(level)(attr.__func__)))
        elif isinstance(attr, staticmethod):
            setattr(cls, attr_name, staticmethod(track_func(level, static=True)(attr.__func__)))
    # return cls(*args, **kwargs)

# def track_class(level):

#     def deco(cls):
#         @functools.wraps(cls)
#         def class_wrapper(*args, **kwargs):
#             return wrap_methods(cls, level, *args, **kwargs)
#         return class_wrapper

#     return deco


def track_class(cls, level):
    wrap_methods(cls, level)
    return cls


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import traceback
from arjuna.tpi.constant import ArjunaOption
from arjuna.interact.gui.auto.finder.wmd import SimpleGuiWidgetMetaData

from arjuna.core.poller.conditions import *
from arjuna.core.poller.caller import *
from arjuna.tpi.error import ChildWindowNotFoundError

class WindowConditions:

    def __init__(self, window):
        self.__window = window

    @property
    def window(self):
        return self.__window

    def ChildWindowIsPresent(self, wmd):
        caller = DynamicCaller(self.window._find_child_window, wmd)
        return CommandCondition(caller)

class BasicWindow:

    def __init__(self, app, automator):
        super().__init__()
        self.__app = app
        self.__automator = automator
        self.__window_handle = None
        self.__config = automator.config

    @property
    def max_wait(self):
        return self.__automator.config.guiauto_max_wait

    @property
    def app(self):
        return self.__app

    @property
    def config(self):
        return self.__config    

    @property
    def _automator(self):
        return self.__automator

    @property
    def handle(self):
        return self.__window_handle

    @property
    def title(self):
        return self.__automator.dispatcher.get_current_window_title()

    @property
    def size(self):
        return self.__automator.dispatcher.get_current_window_size()

    def _set_handle(self, handle):
        self.__window_handle = handle

    def focus(self):
        self.__automator.dispatcher.focus_on_window(self.handle)

    def is_main_window(self):
        return False

    def set_window_size(self, width, height):
        self.__automator.dispatcher.set_current_window_size(width, height)

    def maximize(self):
        self.__automator.dispatcher.maximize_current_window()

class MainWindow(BasicWindow):

    def __init__(self, app, automator):
        super().__init__(app, automator)
        self._set_handle(self.get_current_window_handle()) 
        self.__all_child_windows = {}
        self.__setu_id_map = {}
        self.__resize_window_as_per_config()
        self.__conditions = WindowConditions(self)

    @property
    def conditions(self):
        return self.__conditions

    def get_all_child_window_handles(self):
        handles = self.__automator.dispatcher.get_all_window_handles()
        handles = [handle for handle in handles if handle != self.handle]
        new_handles = []
        for handle in handles:
            if handle != self.handle:
                if handle not in self.__all_child_windows:
                    cwin = ChildWindow(self.app, self.__automator, self, handle)
                    self.__all_child_windows[handle] = cwin
                    new_handles.append(handle)
        return handles, new_handles

    @property
    def latest_child_window(self):
        _, new_handles = self.get_all_child_window_handles()
        if not new_handles:
            raise Exception("No new window was launched.")
        elif len(new_handles) > 1:
            raise Exception("Multiple new windows were launched, so can not deterministically jump to a new window.")
        return self.__all_child_windows[new_handles[0]]

    def __get_child_window(self, handle):
        return self.__all_child_windows[handle]

    def __focus_on_window(self, handle):
        self.__get_child_window(handle).jump()

    def delete_window(self, handle):
        del self.__all_child_windows[handle]

    def close_all_child_windows(self):
        all_child_handles, _ = self.get_all_child_window_handles()
        for handle in all_child_handles:
            cwin = self.__get_child_window(handle)
            cwin.focus()
            cwin.close()
        self.focus()

    def get_current_window_handle(self):
        return self._automator.dispatcher.get_current_window_handle()

    def child_window(self, wmd):
        return self.get_conditions().ChildWindowIsPresent(wmd).wait(max_wait=self.max_wait)

    def _find_child_window(self, wmd):
        all_child_handles, _ = self.get_all_child_window_handles()
        for handle in all_child_handles:
            cwin = self.__get_child_window(handle)
            cwin.focus()
            for locator in wmd.locators:
                if locator.ltype.name == "WINDOW_TITLE":
                    if cwin.title.lower() == locator.lvalue.lower():
                        return cwin
                elif locator.ltype.name == "WINDOW_PTITLE":
                    if locator.lvalue.lower() in cwin.title.lower():
                        return cwin
                elif locator.ltype.name == "CONTENT_LOCATOR":
                    for child_locator in locator.lvalue.locators:
                        try:
                            wmd = SimpleGuiWidgetMetaData(child_locator.ltype.name, child_locator.lvalue)
                            # The element for window is created in the context of an app.
                            # Need to verify this logic.
                            # and its impact on POM.
                            contained_element = self.__automator.element(self.app,wmd, max_wait=0.5)
                            # contained_element.find()
                            return cwin
                        except WaitableError as f:
                            continue  
                        except Exception as e:
                            raise Exception(str(e) + traceback.format_exc())                          
        
        raise ChildWindowNotFoundError(*wmd.locators)

    def __resize_window_as_per_config(self):
        # Resize window
        config = self.config
        browser_width = config.value(ArjunaOption.BROWSER_DIM_WIDTH)
        browser_height = config.value(ArjunaOption.BROWSER_DIM_HEIGHT)
        should_maximize = config.value(ArjunaOption.BROWSER_MAXIMIZE)

        if config.is_arjuna_option_not_set(ArjunaOption.BROWSER_DIM_WIDTH) and config.is_arjuna_option_not_set(ArjunaOption.BROWSER_DIM_HEIGHT):
            if should_maximize:
                self.maximize()
        else:
            width, height = None, None
            current_width, current_height = self.get_size()
            width = config.is_arjuna_option_not_set(ArjunaOption.BROWSER_DIM_WIDTH) and browser_width or current_width
            height = config.is_arjuna_option_not_set(ArjunaOption.BROWSER_DIM_HEIGHT) and browser_height or current_height
            self.set_window_size(width, height)

    def is_main_window(self):
        return True

    def close(self):
        raise Exception("You can not close main window. Use automator.quit() to quit application.")

class ChildWindow(BasicWindow):

    def __init__(self, app, automator, main_window, handle):
        super().__init__(app, automator)
        self.__main_window = main_window
        self._set_handle(handle) 

    def close(self):
        self.focus()
        self._automator.dispatcher.close_current_window()
        self.__main_window.delete_window(self.handle)
        self.__main_window.focus()


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import os
import copy
import abc
from collections import namedtuple
from lxml import etree
from arjuna.tpi.parser.html import Html, HtmlNode
from arjuna.tpi.tracker import track
from .content import GuiSourceContent

@track("trace")
class GuiSource:
    '''
        GUI Source of a GUI entity.

        Not meant to be directly constructed by a test author. Retrieve as:
        
            .. code-block:: python
                
                gui_entity.source
    '''
    def __init__(self):
        self.__content = None

    @property
    def content(self) -> GuiSourceContent:
        '''
            Source content of associated GUI entity returned as :class:`~arjuna.tpi.guiauto.source.content.GuiSourceContent`.
        '''
        return self.__content

    @content.setter
    def _content(self, content):
        self.__content = content

@track("trace")
class SingleGuiEntitySource(GuiSource, metaclass=abc.ABCMeta):
    '''
        Abstract Base class for GUI Source of a Singular GUI entity (**GuiPage** or **GuiElement**).

        Not meant to be directly constructed by a test author. Retrieve as:
        
            .. code-block:: python
                
                gui_entity.source

        Args:
            raw_source: Raw XML source.
            root_tag: 'html' for Gui and 'body' for GuiElement.
    '''

    def __init__(self, raw_source, partial=False):
        super().__init__()
        self.__raw_source = raw_source
        self.__partial = partial
        self.__fpaths = []
        self.__node = None 
        self.__elem_node = None

    @property
    def node(self) -> HtmlNode:
        '''
            Source code as an Arjuna :class:`~arjuna.tpi.parser.xml.XmlNode` for advanced inquiry and parsing.
        '''
        return self.__node

    @property
    def _elem_node(self):
        return self.__elem_node

    def _process_elem_node(self, elem_node):
        pass

    def _load(self):
        raw_source = self.__raw_source
        self.__elem_node = Html.from_str(raw_source, partial=self.__partial)
        self.__node = self.__elem_node.clone()
        self.__text_content = self.__elem_node.normalized_text
        self.__inner_html = self.__elem_node.normalized_inner_html
        self.__full_source = raw_source
        self.__elem_node.remove_all_children()
        self.__self_source = self.__elem_node.normalized_source
        self._content = GuiSourceContent(all=self.__full_source, root=self.__self_source, inner=self.__inner_html, text=self.__text_content)
        self._process_elem_node(self.__elem_node)



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# from arjuna.tpi.constant import ArjunaOption
# from arjuna.client.core.action import *
# from arjuna.interact.gui.auto.finder.wmd import SimpleGuiWidgetMetaData

# from .guiautomator import GuiAutomator
# from .handler import Handler

# class SingleActionChain(Handler):

#     def __init__(self, automator, element=None):
#         super().__init__(automator)
#         from arjuna.interact.gui.auto.widget.frame import DomRoot
#         self.__automator = automator
#         self.__element = element
#         self.__attached_to_element = element is not None

#     def is_click_action(self, action):
#         return PartialActionType[action["actionType"]] == PartialActionType.CLICK

#     def are_args_defined(self, action):
#         return "args" in action

#     def get_args(self, action):
#         return action["args"]

#     def contains_target_element(self, action):
#         return "targetElement" in self.get_args(action)

#     def get_target_element(self, action):
#         return action["args"]["targetElement"]

#     def contains_target_point(self, action):
#         return "targetPoint" in self.get_args(action)

#     def get_target_point(self, action):
#         return action["args"]["targetPoint"]

#     def perform(self, single_action_chain):
#         processed_list = []
#         current_action_list = []
#         for action in single_action_chain:
#             if self.is_click_action(action):
#                 if not self.are_args_defined():
#                     current_action_list.append(("click", dict()))
#                     continue

#                 target_element = None
#                 if self.contains_target_element(action):
#                     target_element_id = self.get_target_element(action)
#                     target_element = self.__automator.get_element_for_setu_id(target_element_id)
#                 elif self.contains_target_point(action):
#                     wmd = SimpleGuiWidgetMetaData("point", self.get_target_point(action))
#                     element = self.__automator.create_element(wmd)
                
#                 if target_element:
#                     target_element.find_if_not_found()
#                     current_action_list.append(("click", {"on_element" : (target_element.setu_id, True)}))
                    
#         processed_list.append(current_action_list)
#         for single_chain in processed_list:
#             self.__automator.dispatcher.perform_action_chain(single_chain)


        



